package sun.nio.cs;

import java.nio.charset.CoderResult;
import java.nio.CharBuffer;
import java.nio.charset.CodingErrorAction;
import java.nio.channels.FileChannel;
import java.io.FileInputStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.IllegalCharsetNameException;
import java.io.IOException;
import java.nio.channels.ReadableByteChannel;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.Charset;
import java.io.Reader;

public class StreamDecoder extends Reader
{
    private static final int MIN_BYTE_BUFFER_SIZE = 32;
    private static final int DEFAULT_BYTE_BUFFER_SIZE = 8192;
    private volatile boolean isOpen;
    private boolean haveLeftoverChar;
    private char leftoverChar;
    private static volatile boolean channelsAvailable;
    private Charset cs;
    private CharsetDecoder decoder;
    private ByteBuffer bb;
    private InputStream in;
    private ReadableByteChannel ch;
    
    private void ensureOpen() throws IOException {
        if (!this.isOpen) {
            throw new IOException("Stream closed");
        }
        return;
    }
    
    public static StreamDecoder forInputStreamReader(InputStream inputStream, Object o, String name) throws UnsupportedEncodingException {
        if (name == null) {
            name = Charset.defaultCharset().name();
        }
        try {
            if (Charset.isSupported(name)) {
                return new StreamDecoder(inputStream, o, Charset.forName(name));
            }
        }
        catch (IllegalCharsetNameException ex) {}
        throw new UnsupportedEncodingException(name);
    }
    
    public static StreamDecoder forInputStreamReader(InputStream inputStream, Object o, Charset charset) {
        return new StreamDecoder(inputStream, o, charset);
    }
    
    public static StreamDecoder forInputStreamReader(InputStream inputStream, Object o, CharsetDecoder charsetDecoder) {
        return new StreamDecoder(inputStream, o, charsetDecoder);
    }
    
    public static StreamDecoder forDecoder(ReadableByteChannel readableByteChannel, CharsetDecoder charsetDecoder, int n) {
        return new StreamDecoder(readableByteChannel, charsetDecoder, n);
    }
    
    public String getEncoding() {
        if (this.isOpen()) {
            return this.encodingName();
        }
        return null;
    }
    
    @Override
    public int read() throws IOException {
        return this.read0();
    }
    
    private int read0() throws IOException {
        synchronized (this.lock) {
            if (this.haveLeftoverChar) {
                this.haveLeftoverChar = false;
                return this.leftoverChar;
            }
            char[] array = new char[2];
            int read = this.read(array, 0, 2);
            switch (read) {
                case -1: {
                    return -1;
                }
                case 2: {
                    this.leftoverChar = array[1];
                    this.haveLeftoverChar = true;
                }
                case 1: {
                    return array[0];
                }
                default: {
                    assert false : read;
                    return -1;
                }
            }
        }
    }
    
    @Override
    public int read(char[] array, int n, int n2) throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            if (n < 0 || n > array.length || n2 < 0 || n + n2 > array.length || n + n2 < 0) {
                throw new IndexOutOfBoundsException();
            }
            if (n2 == 0) {
                return 0;
            }
            int n3 = 0;
            if (this.haveLeftoverChar) {
                array[n] = this.leftoverChar;
                ++n;
                --n2;
                this.haveLeftoverChar = false;
                n3 = 1;
                if (n2 == 0 || !this.implReady()) {
                    return n3;
                }
            }
            if (n2 == 1) {
                int read0 = this.read0();
                if (read0 == -1) {
                    return (n3 == 0) ? -1 : n3;
                }
                array[n] = (char)read0;
                return n3 + 1;
            }
            else {
                return n3 + this.implRead(array, n, n + n2);
            }
        }
    }
    
    @Override
    public boolean ready() throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            return this.haveLeftoverChar || this.implReady();
        }
    }
    
    @Override
    public void close() throws IOException {
        synchronized (this.lock) {
            if (!this.isOpen) {
                return;
            }
            this.implClose();
            this.isOpen = false;
        }
    }
    
    private boolean isOpen() {
        return this.isOpen;
    }
    
    private static FileChannel getChannel(FileInputStream fileInputStream) {
        if (!StreamDecoder.channelsAvailable) {
            return null;
        }
        try {
            return fileInputStream.getChannel();
        }
        catch (UnsatisfiedLinkError unsatisfiedLinkError) {
            StreamDecoder.channelsAvailable = false;
            return null;
        }
    }
    
    StreamDecoder(InputStream inputStream, Object o, Charset charset) {
        this(inputStream, o, charset.newDecoder().onMalformedInput(CodingErrorAction.REPLACE).onUnmappableCharacter(CodingErrorAction.REPLACE));
    }
    
    StreamDecoder(InputStream in, Object lock, CharsetDecoder decoder) {
        super(lock);
        this.isOpen = true;
        this.haveLeftoverChar = false;
        this.cs = decoder.charset();
        this.decoder = decoder;
        if (this.ch == null) {
            this.in = in;
            this.ch = null;
            this.bb = ByteBuffer.allocate(8192);
        }
        this.bb.flip();
    }
    
    StreamDecoder(ReadableByteChannel ch, CharsetDecoder decoder, int n) {
        this.isOpen = true;
        this.haveLeftoverChar = false;
        this.in = null;
        this.ch = ch;
        this.decoder = decoder;
        this.cs = decoder.charset();
        (this.bb = ByteBuffer.allocate((n < 0) ? 8192 : ((n < 32) ? 32 : n))).flip();
    }
    
    private int readBytes() throws IOException {
        this.bb.compact();
        try {
            if (this.ch != null) {
                int read2 = this.ch.read(this.bb);
                if (read2 < 0) {
                    return read2;
                }
            }
            else {
                int limit = this.bb.limit();
                int position = this.bb.position();
                assert position <= limit;
                int n = (position <= limit) ? (limit - position) : 0;
                assert n > 0;
                int read = this.in.read(this.bb.array(), this.bb.arrayOffset() + position, n);
                if (read < 0) {
                    return read;
                }
                if (read == 0) {
                    throw new IOException("Underlying input stream returned zero bytes");
                }
                assert read <= n : "n = " + read + ", rem = " + n;
                this.bb.position(position + read);
            }
        }
        finally {
            this.bb.flip();
        }
        int remaining = this.bb.remaining();
        assert remaining != 0 : remaining;
        return remaining;
    }
    
    int implRead(char[] array, int offset, int n) throws IOException {
        assert n - offset > 1;
        CharBuffer out = CharBuffer.wrap(array, offset, n - offset);
        if (out.position() != 0) {
            out = out.slice();
        }
        boolean endOfInput = false;
        while (true) {
            CoderResult decode = this.decoder.decode(this.bb, out, (boolean)endOfInput);
            if (decode.isUnderflow()) {
                if (!(endOfInput == 0)) {
                    break;
                }
                if (!out.hasRemaining()) {
                    break;
                }
                if (out.position() > 0 && !this.inReady()) {
                    break;
                }
                if (this.readBytes() >= 0) {
                    continue;
                }
                endOfInput = true;
                if (out.position() == 0 && !this.bb.hasRemaining()) {
                    break;
                }
                this.decoder.reset();
            }
            else if (decode.isOverflow()) {
                assert out.position() > 0;
                break;
            }
            else {
                decode.throwException();
            }
        }
        if (!(endOfInput == 0)) {
            this.decoder.reset();
        }
        if (out.position() == 0) {
            if (!(endOfInput == 0)) {
                return -1;
            }
            assert false;
        }
        return out.position();
    }
    
    String encodingName() {
        return (this.cs instanceof HistoricallyNamedCharset) ? ((HistoricallyNamedCharset)this.cs).historicalName() : this.cs.name();
    }
    
    private boolean inReady() {
        try {
            return (this.in != null && this.in.available() > 0) || this.ch instanceof FileChannel;
        }
        catch (IOException ex) {
            return false;
        }
    }
    
    boolean implReady() {
        return this.bb.hasRemaining() || this.inReady();
    }
    
    void implClose() throws IOException {
        if (this.ch != null) {
            this.ch.close();
        }
        else {
            this.in.close();
        }
    }
    
    static {
        StreamDecoder.channelsAvailable = true;
    }
}
